"""
Module for reporting into http://www.blazemeter.com/ service

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from bzt import TaurusConfigError
from bzt.bza import Test
from bzt.modules.blazemeter.cloud_test import CloudTaurusTest
from bzt.modules.blazemeter.net_utils import parse_blazemeter_test_link


class ProjectFinder(object):
    def __init__(self, parameters, settings, user, workspaces, parent_log):
        super(ProjectFinder, self).__init__()
        self.default_test_name = "Taurus Test"
        self.parameters = parameters
        self.settings = settings
        self.log = parent_log.getChild(self.__class__.__name__)
        self.user = user
        self.workspaces = workspaces
        self.test_type = None

    def _find_project(self, proj_name):
        """
        :rtype: bzt.bza.Project
        """
        if isinstance(proj_name, (int, float)):  # project id
            proj_id = int(proj_name)
            self.log.debug("Treating project name as ID: %s", proj_id)
            project = self.workspaces.projects(ident=proj_id).first()
            if not project:
                raise TaurusConfigError("BlazeMeter project not found by ID: %s" % proj_id)
        elif proj_name:
            project = self.workspaces.projects(name=proj_name).first()
        else:
            project = None

        return project

    def _ws_proj_switch(self, project):
        if project:
            return project
        else:
            return self.workspaces

    def resolve_external_test(self):
        proj_name = self.parameters.get("project", self.settings.get("project"))
        test_name = self.parameters.get("test", self.settings.get("test", self.default_test_name))

        project = self._find_project(proj_name)
        if not project and proj_name:
            project = self.workspaces.first().create_project(proj_name)

        test = self._ws_proj_switch(project).tests(name=test_name, test_type='external').first()

        if not test:
            if not project:
                info = self.user.fetch()
                project = self.workspaces.projects(ident=info['defaultProject']['id']).first()
                if not project:
                    project = self.workspaces.first().create_project("Taurus Tests Project")

            test = project.create_test(test_name, {"type": "external"})

        return test

    def resolve_account(self, account_name):
        account = None

        if isinstance(account_name, (int, float)):
            acc_id = int(account_name)
            self.log.debug("Treating account name as ID: %s", acc_id)
            account = self.user.accounts(ident=acc_id).first()
            if not account:
                raise TaurusConfigError("BlazeMeter account not found by ID: %s" % acc_id)
        elif account_name:
            account = self.user.accounts(name=account_name).first()
            if not account:
                raise TaurusConfigError("BlazeMeter account not found by name: %s" % account_name)

        if account:
            return account

        self.user.fetch()
        account = self.user.accounts(ident=self.user['defaultProject']['accountId']).first()
        self.log.debug("Using default account: %s", account)
        return account

    def resolve_workspace(self, account, workspace_name):
        workspace = None

        if isinstance(workspace_name, (int, float)):
            workspace_id = int(workspace_name)
            self.log.debug("Treating workspace name as ID: %s", workspace_id)
            workspace = account.workspaces(ident=workspace_id).first()
            if not workspace:
                raise TaurusConfigError("BlazeMeter workspace not found by ID: %s" % workspace_id)
        elif workspace_name is not None:
            workspace = account.workspaces(name=workspace_name).first()
            if not workspace:
                raise TaurusConfigError("BlazeMeter workspace not found: %s" % workspace_name)

        if workspace is None:
            workspace = account.workspaces(ident=self.user['defaultProject']['workspaceId']).first()
            self.log.debug("Using first workspace: %s" % workspace)

        return workspace

    def resolve_project(self, workspace, project_name):
        if isinstance(project_name, (int, float)):  # project id
            project_id = int(project_name)
            self.log.debug("Treating project name as ID: %s", project_id)
            project = workspace.projects(ident=project_id).first()
            if not project:
                raise TaurusConfigError("BlazeMeter project not found by ID: %s" % project_id)
        elif project_name:
            project = workspace.projects(name=project_name).first()
        else:
            project = None

        if not project:
            project = self._create_project_or_use_default(workspace, project_name)

        return project

    def _create_project_or_use_default(self, workspace, proj_name):
        if proj_name:
            return workspace.create_project(proj_name)
        else:
            info = self.user.fetch()
            self.log.debug("Looking for default project: %s", info['defaultProject']['id'])
            project = self.workspaces.projects(ident=info['defaultProject']['id']).first()
            if not project:
                project = workspace.create_project("Taurus Tests Project")
            return project

    def resolve_test(self, project, test_name, test_type):
        is_int = isinstance(test_name, (int, float))
        is_digit = isinstance(test_name, str) and test_name.isdigit()

        if is_int or is_digit:
            test_id = int(test_name)
            self.log.debug("Treating project name as ID: %s", test_id)
            test = project.tests(ident=test_id, test_type=test_type).first()
            if not test:
                raise TaurusConfigError("BlazeMeter test not found by ID: %s" % test_id)
        elif test_name is not None:
            test = project.tests(name=test_name, test_type=test_type).first()
        else:
            test = None

        return test

    def get_test_router(self):
        default_location = self.settings.get("default-location", None)
        account_name = self.parameters.get("account", self.settings.get("account", None))
        workspace_name = self.parameters.get("workspace", self.settings.get("workspace", None))
        project_name = self.parameters.get("project", self.settings.get("project"))
        test_name = self.parameters.get("test", self.settings.get("test", self.default_test_name))
        launch_existing_test = self.settings.get("launch-existing-test", False)

        # if we're to launch existing test - don't use test_type filter (look for any type)
        filter_test_type = None if launch_existing_test else self.test_type

        test_spec = parse_blazemeter_test_link(test_name)
        self.log.debug("Parsed test link: %s", test_spec)
        if test_spec is not None:
            account, workspace, project, test = self.user.test_by_ids(test_spec.account_id, test_spec.workspace_id,
                                                                      test_spec.project_id, test_spec.test_id,
                                                                      test_type=filter_test_type)
            if not test:
                raise TaurusConfigError("Test not found: %s", test_name)
            self.log.info("Found test by link: %s", test_name)
        else:
            account = self.resolve_account(account_name)
            workspace = self.resolve_workspace(account, workspace_name)
            project = self.resolve_project(workspace, project_name)
            test = self.resolve_test(project, test_name, test_type=filter_test_type)

        if isinstance(test, Test):
            test_class = CloudTaurusTest
        else:   # test not found
            if launch_existing_test:
                raise TaurusConfigError("Can't find test for launching: %r" % test_name)

            test_class = CloudTaurusTest

        router = test_class(self.user, test, project, test_name, default_location, launch_existing_test, self.log)
        router._workspaces = self.workspaces
        router.cloud_mode = self.settings.get("cloud-mode", None)
        router.dedicated_ips = self.settings.get("dedicated-ips", False)
        router.test_type = self.test_type
        router.send_report_email = self.settings.get("send-report-email", False)
        return router



from selenium.webdriver.remote.webelement import WebElement


class ShadowElement(WebElement):
    """
    This class extends WebElement to allow for overriding methods that do not work properly in the Shadow components
    Currently the issues are mostly on salesforce site and so far only related to click action
    """

    def __init__(self, element, driver):
        self.element = element
        self.driver = driver

    def __getattr__(self, attr):
        return getattr(self.element, attr)

    def click(self):
        self.driver.execute_script("arguments[0].click();", self.element)



import json
import sys
import time
from datetime import datetime
from optparse import OptionParser

from apiritif.samples import Sample
from io import StringIO
from robot import run


class TaurusListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, report_filename):
        self._report_filename = report_filename
        self._report_file = None
        self._test_count = 0
        self._passed_tests = 0
        self._failed_tests = 0
        self._current_suite = None

    def prepare(self):
        self._report_file = open(self._report_filename, 'wt')

    def post_process(self):
        if self._report_file is not None:
            self._report_file.close()

    def get_test_count(self):
        return self._test_count

    def _report_sample(self, sample):
        self._test_count += 1
        if sample.status == "PASSED":
            self._passed_tests += 1
        elif sample.status in ["BROKEN", "FAILED"]:
            self._failed_tests += 1

        self._write_sample(sample)
        self._write_stdout_report(sample.test_case)

    def _write_sample(self, sample):
        if self._report_file is None:
            raise ValueError("Plugin wasn't prepared")

        self._report_file.write("%s\n" % json.dumps(sample.to_dict()))
        self._report_file.flush()

    def _write_stdout_report(self, label):
        report_pattern = "%s,Total:%d Passed:%d Failed:%d\n"
        sys.stdout.write(report_pattern % (label, self._test_count, self._passed_tests, self._failed_tests))
        sys.stdout.flush()

    def start_suite(self, name, attrs):
        self._current_suite = name

    def start_test(self, name, attrs):
        pass

    def end_test(self, name, attrs):
        # TODO: include keywords as subsamples
        sample = Sample()
        sample.test_case = name
        sample.test_suite = self._current_suite
        sample.start_time = time.mktime(datetime.strptime(attrs['starttime'], '%Y%m%d %H:%M:%S.%f').timetuple())
        sample.duration = float(attrs['elapsedtime']) / 1000.0
        if attrs['status'] == 'PASS':
            sample.status = 'PASSED'
        else:
            sample.status = 'FAILED'
            message = attrs['message']
            if '\n' in message:
                lines = message.split('\n')
                sample.error_msg = lines[0]
                sample.error_trace = message
            else:
                sample.error_msg = message
        self._report_sample(sample)

    def end_suite(self, name, attrs):
        self._current_suite = None


def run_robot(targets, report_file, iteration_limit, duration_limit, variablefile, outputfile, logfile, include, cmdline):
    listener = TaurusListener(report_file)
    listener.prepare()
    stdout = StringIO()
    stderr = StringIO()
    start_time = int(time.time())
    iteration = 0
    try:
        while True:
            kwargs = {
                'listener': listener,  # pass Taurus listener
                'output': outputfile, 'log': logfile, 'report': None,
                'stdout': stdout, 'stderr': stderr,  # capture stdout/stderr
            }
            if variablefile is not None:
                kwargs['variablefile'] = variablefile
            if include is not None:
                kwargs['include'] = include
            if cmdline:
                parts = cmdline[1:-1].split(" ")
                part = iter(parts)
                kwargs.update(dict(zip(part, part)))

            run(*targets, **kwargs)
            iteration += 1
            if 0 < duration_limit < int(time.time()) - start_time:
                break
            if iteration >= iteration_limit:
                break
    finally:
        listener.post_process()

    sys.stdout.write("Robot stdout:\n" + stdout.getvalue() + "\n")
    sys.stderr.write("Robot stderr:\n" + stderr.getvalue() + "\n")

    if listener.get_test_count() == 0:
        raise ValueError("Nothing to test. No tests were found.")


if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option('-r', '--report-file', action='store', default='report.ldjson')
    parser.add_option('-i', '--iterations', action='store', default=0)
    parser.add_option('-d', '--duration', action='store', default=0)
    parser.add_option('-v', '--variablefile', action='store', default=None)
    parser.add_option('-o', '--outputfile', action='store', default=None)
    parser.add_option('-l', '--logfile', action='store', default=None)
    parser.add_option('--include', action='store', default=None)
    parser.add_option('--cmdline', action='store', default=None)
    opts, args = parser.parse_args()
    if opts.include is not None:
        opts.include = opts.include.split(',')
    opts.iterations = int(opts.iterations)
    opts.duration = float(opts.duration)

    if opts.iterations == 0:
        if opts.duration > 0:
            opts.iterations = sys.maxsize
        else:
            opts.iterations = 1

    run_robot(args, opts.report_file, int(opts.iterations), float(opts.duration), opts.variablefile,
              opts.outputfile, opts.logfile, opts.include, opts.cmdline)



"""
Copyright 2018 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import multiprocessing
import os
import re
import shlex
import sys

from bzt import TaurusConfigError
from bzt.engine import SETTINGS, Provisioning
from bzt.modules import SubprocessedExecutor, ScenarioExecutor
from bzt.modules.services import PythonTool
from bzt.utils import FileReader, RESOURCES_DIR, RequiredTool

IGNORED_LINE = re.compile(r"[^,]+,Total:\d+ Passed:\d+ Failed:\d+")


class PyTestExecutor(SubprocessedExecutor):
    def __init__(self):
        super(PyTestExecutor, self).__init__()
        self.runner_path = os.path.join(RESOURCES_DIR, "pytest_runner.py")
        self._tailer = FileReader('', file_opener=lambda _: None, parent_logger=self.log)
        self._additional_args = []
        self.pytest = None

    def prepare(self):
        super(PyTestExecutor, self).prepare()
        self.install_required_tools()
        self.script = self.get_script_path()
        if not self.script:
            raise TaurusConfigError("'script' should be present for pytest executor")

        scenario = self.get_scenario()
        if "additional-args" in scenario:
            argv = scenario.get("additional-args")
            self._additional_args = shlex.split(argv)

        self.reporting_setup(suffix=".ldjson")

    def __is_verbose(self):
        engine_verbose = self.engine.config.get(SETTINGS).get("verbose", False)
        executor_verbose = self.settings.get("verbose", engine_verbose)
        return executor_verbose

    def install_required_tools(self):
        """
        we need installed nose plugin
        """
        self.pytest = self._get_tool(PyTest, engine=self.engine, settings=self.settings)
        self._check_tools([self.pytest, self._get_tool(TaurusPytestRunner, tool_path=self.runner_path)])

    def get_load(self):
        raw_concurrency = str(self.get_raw_load().concurrency).lower()
        if raw_concurrency == 'auto':
            prov_type = self.engine.config.get(Provisioning.PROV)
            self.execution.get(ScenarioExecutor.CONCURR)[prov_type] = -1
        return super().get_load()

    def startup(self):
        """
        run python tests
        """
        executable = self.settings.get("interpreter", sys.executable)

        cmdline = [executable, self.runner_path, '--report-file', self.report_file]

        load = self.get_load()
        if load.iterations:
            cmdline += ['-i', str(load.iterations)]

        if load.hold:
            cmdline += ['-d', str(load.hold)]

        raw_concurrency = self.get_raw_load().concurrency

        if raw_concurrency:
            if raw_concurrency == -1:
                raw_concurrency = 'auto'
            elif raw_concurrency > multiprocessing.cpu_count():
                raw_concurrency = multiprocessing.cpu_count()
                self.log.warning(f"pytest concurrency is limited to CPU count [{raw_concurrency}]")
            cmdline.extend(['-n', str(raw_concurrency)])

        cmdline += self._additional_args
        cmdline += [self.script]

        self.process = self._execute(cmdline)

        if self.__is_verbose():
            self._tailer = FileReader(filename=self.stdout.name, parent_logger=self.log)

    def check(self):
        self.__log_lines()
        return super(PyTestExecutor, self).check()

    def post_process(self):
        self.pytest.post_process()
        super(PyTestExecutor, self).post_process()
        self.__log_lines()

    def __log_lines(self):
        lines = []
        for line in self._tailer.get_lines():
            if not IGNORED_LINE.match(line):
                lines.append(line)

        if lines:
            self.log.info("\n".join(lines))


class PyTest(PythonTool):
    PACKAGES = ["pytest", "pytest-xdist", "apiritif"]


class TaurusPytestRunner(RequiredTool):
    def __init__(self, tool_path, **kwargs):
        super(TaurusPytestRunner, self).__init__(tool_path=tool_path, installable=False, **kwargs)



"""
Copyright 2018 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys

import yaml

from bzt import TaurusConfigError
from bzt.modules import SubprocessedExecutor
from bzt.modules.services import PythonTool
from bzt.utils import RequiredTool, CALL_PROBLEMS
from bzt.utils import get_full_path, RESOURCES_DIR


class RobotExecutor(SubprocessedExecutor):
    def __init__(self):
        super(RobotExecutor, self).__init__()
        self.runner_path = os.path.join(RESOURCES_DIR, "robot_runner.py")
        self.variables_file = None
        self.output_file = None
        self.log_file = None
        self.tags = None
        self.robot = None

    def resource_files(self):
        files = super(RobotExecutor, self).resource_files()
        scenario = self.get_scenario()
        if "variables" in scenario and isinstance(scenario["variables"], str):
            files.append(scenario["variables"])
        return files

    def prepare(self):
        super(RobotExecutor, self).prepare()
        self.install_required_tools()
        self.script = self.get_script_path()
        if not self.script:
            raise TaurusConfigError("'script' should be present for robot executor")

        self.reporting_setup(suffix=".ldjson")
        self.output_file = self.engine.create_artifact("output", ".xml")
        self.log_file = self.engine.create_artifact("log", ".html")

        scenario = self.get_scenario()
        variables = scenario.get("variables")
        if variables:
            if isinstance(variables, str):
                self.variables_file = get_full_path(variables)
            elif isinstance(variables, dict):
                self.variables_file = self.engine.create_artifact("robot-vars", ".yaml")
                with open(self.variables_file, 'wb') as fds:
                    yml = yaml.safe_dump(variables,
                                         default_flow_style=False, explicit_start=True, canonical=False,
                                         allow_unicode=True,
                                         encoding='utf-8', width=float("inf"))
                    fds.write(yml)
            else:
                raise TaurusConfigError("`variables` is neither file nor dict")
        tags = scenario.get("tags", None)
        if tags:
            if isinstance(tags, str):
                self.tags = tags
            else:
                raise TaurusConfigError("`tags` is not a string or text")

    def install_required_tools(self):
        self.robot = self._get_tool(Robot, engine=self.engine, settings=self.settings)
        self._check_tools([self.robot, self._get_tool(TaurusRobotRunner, tool_path=self.runner_path)])

    def startup(self):
        executable = self.settings.get("interpreter", sys.executable)

        cmdline = [executable, self.runner_path, '--report-file', self.report_file]

        load = self.get_load()
        if load.iterations:
            cmdline += ['--iterations', str(load.iterations)]

        if load.hold:
            cmdline += ['--duration', str(load.hold)]

        if self.variables_file is not None:
            cmdline += ['--variablefile', self.variables_file]

        if self.tags is not None:
            cmdline += ['--include', self.tags]

        cmdline += ['--outputfile', self.output_file]
        cmdline += ['--logfile', self.log_file]

        user_cmd = self.settings.get("cmdline")
        if user_cmd:
            cmdline.append("--cmdline")
            cmdline.append(f'"{user_cmd.replace("=", " ").replace("-", "")}"')

        cmdline += [self.script]
        self.process = self._execute(cmdline)

    def post_process(self):
        self.robot.post_process()
        super(RobotExecutor, self).post_process()


class Robot(PythonTool):
    PACKAGES = ["robotframework", "apiritif", "robotframework-seleniumlibrary"]


class TaurusRobotRunner(RequiredTool):
    def __init__(self, tool_path, **kwargs):
        super(TaurusRobotRunner, self).__init__(tool_path=tool_path, installable=False, **kwargs)



from .names import EXEC, TAURUS_ARTIFACTS_DIR, SETTINGS
from .engine import Engine
from .dicts import Scenario, Configuration
from .templates import HavingInstallableTools, Singletone
from .modules import Provisioning, Reporter, Service, Aggregator, EngineModule, ScenarioExecutor
